(* automate du tas de sable, selon Dhar *)

let visuel = ref false;;
let f = ref 1;;

#open "graphics";;

open_graph "";;

let colorie  = function 
 | 0 -> white
 | 1 -> yellow
 | 2 -> green
 | 3 -> blue
 | 4 -> red
 | _ -> black;;

let pixel i j c = 
 set_color c;
 fill_rect (!f*i) (!f*j) !f !f;;

let dessine m = 
 clear_graph();
 for i = 0 to vect_length m - 1 do 
  for j = 0 to vect_length m.(0) - 1 do
    pixel i j (colorie m.(i).(j))
  done
 done;;

(* incrmente un site *)

let incremente m i j d = 
 try 
  m.(i).(j) <- m.(i).(j) + d
 with _ -> ();;

(* boule un site de hauteur au moins 4 *)
 
let eboule m i j = 
 incremente m i j (-4); 
 incremente m (i-1) j 1;
 incremente m (i+1) j 1;
 incremente m i (j-1) 1;
 incremente m i (j+1) 1;;

(* dresse la liste des sites instables *)

let instable x = x>3;;

let liste_sites_instables m = 
 let p = vect_length m and q = vect_length m.(0) in 
 let rec aux i j accu = 
  if i=p then accu else 
  if j=q then aux (i+1) 0 accu else 
  aux i (j+1) (if instable m.(i).(j) then (i,j)::accu else accu) 
 in aux 0 0 [];;

let eboule_sites_instables m lsi = 
 let rec aux = function 
  | [] -> () 
  | (i,j)::q -> eboule m i j; aux q 
 in aux lsi;;

(* relaxe un automate *)

let rec relaxe m = 
 let lsi = liste_sites_instables m in match lsi with 
  | [] -> () 
  | _  -> eboule_sites_instables m lsi; relaxe m;;

(* incrmentation des bords, selon Creutz *)

let inc_bord m = 
 let p = vect_length m and q = vect_length m.(0) in 
 for i = 0 to p-1 do 
  incremente m i 0 1;
  incremente m i (q-1) 1 
 done;
 for j = 0 to q-1 do
  incremente m 0 j 1;
  incremente m (p-1) j 1
 done;;

(* double un etat *)

let double m  = 
 let p = vect_length m and q = vect_length m.(0) in 
  for i = 0 to p-1 do 
   for j = 0 to q-1 do
    m.(i).(j) <- 2 * m.(i).(j) 
   done
  done;;

(* copie un etat *)

let copie m m' = 
 let p = vect_length m and q = vect_length m.(0) in 
  for i = 0 to p-1 do 
   for j = 0 to q-1 do
    m'.(i).(j) <- m.(i).(j) 
   done
  done;;

(* ajoute l'tat m  l'tat m' *)

let ajoute m m' = 
 let p = vect_length m and q = vect_length m.(0) in 
  for i = 0 to p-1 do 
   for j = 0 to q-1 do
    m'.(i).(j) <- m'.(i).(j) + m.(i).(j) 
   done
  done;;

(* calcule l'etat identit *)

let identite p q = 
 let  m = make_matrix p q 0 
 and m' = make_matrix p q 0 in 
  inc_bord m;
  let rec itere n = 
   copie m m';
   double m;
   relaxe m;
   if !visuel then dessine m else ();
   if m = m' then (m,n) else itere (n+1) 
  in itere 1;; 

let complemente m = let p = vect_length m and q = vect_length m.(0) in 
 for i=0 to p-1 do for j=0 to q-1 do 
  m.(i).(j) <- 3 - m.(i).(j)
 done done;;

let identite p q = 
 let  m = make_matrix p q 3 
 and m' = make_matrix p q 3 in 
  double m'; relaxe m'; complemente m';
  ajoute m m'; relaxe m'; 
  dessine m';;

let affiche m = 
 let p = vect_length m and q = vect_length m.(0) in 
  for i = 0 to p-1 do 
   for j = 0 to q-1 do
    print_int m.(i).(j) 
   done;
   print_newline()
  done;;

let test p q = 
 let m = identite p q in 
  f := min (size_x()/p) (size_y()/q);
  clear_graph();
  let (m,n) = identite p q in 
   affiche m; 
   print_string "quilibre atteint aprs "; print_int n;
   print_string " itrations.";print_newline();;

(*
Pour modliser certains phnomnes physiques, relevant de ce qu'il est 
convenu d'appeler l'auto-organisation critique, Bak, Tang et Wiesenfeld 
ont introduit en 1987 un automate cellulaire particulier appel 
automate du tas de sable. 

Peu aprs (90-95), Deepak Dhar a montr que cet automate possde 
des proprits algbriques intressantes, ainsi en particulier 
un sous-ensemble de l'espace des configurations (appeles
configurations rcurrentes) peut tre muni d'une structure de groupe ablien. 

Je me propose de montrer de faon lmentaire les rsultats de Dhar. 
Ceux-ci permettent d'introduire un nouveau paramtre associ  chaque graphe, 
ce paramtre est li  ses arbres recouvrants. Dans le cas du graphe complet 
on retrouve ce que certains combinatoristes appellent les fonctions de parking.
*)
 
